#include <dirent.h>
#include <ctype.h>

#include "bluetoothchannel.h"
#define MAX_CONNECTIONS	1

//#define DEBUG_BLUETOOTHCHANNEL

#ifdef DEBUG_BLUETOOTHCHANNEL
  #define BTC_DEBUG(msg...)     do {          printf(msg); fflush(stdout); }   while (0)
  #define BTC_DEBUGT(t, msg...) do { if (t) { printf(msg); fflush(stdout); } } while (0)
#else
  #define BTC_DEBUG(msg...)     do { } while (0)
  #define BTC_DEBUGT(t, msg...) do { } while (0)
#endif

//if you want to have asserts on 0 -> 1
#if 0
  #include <assert.h>
  #define BT_ASSERT(x) do{ assert(x); } while(0)
#else
  #define BT_ASSERT(x) do { } while (0)
#endif

BluetoothChannel::BluetoothChannel()
{
	sprintf( btAddr, "" );
	btChannel = -1;
	sprintf( btTTY, "" );
	btDev = -1;
	btFd = -1;
}

int BluetoothChannel::Initialize(char* aAddr, int aChannel)
{
	//initialize device info
	strcpy( btAddr, aAddr );
	btChannel = aChannel;
	sprintf( btTTY, "" );
	btDev = -1;
	btFd = -1;
}

int BluetoothChannel::Open(volatile int& aCancel)
{
	struct sockaddr_rc remote_addr, local_addr;
	bdaddr_t bdaddr;
	fd_set wfds;
	struct timeval tv;
	int r, s;

	str2ba(btAddr, &bdaddr);
	if ((s = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0)
	{
		BTC_DEBUG("BlueToothChannel::Open(): Can't create socket. %s (%d)\n", strerror(errno), errno);
		return -1;
	}

	memset(&local_addr, 0, sizeof(local_addr));
	local_addr.rc_family = AF_BLUETOOTH;
	bacpy(&local_addr.rc_bdaddr, BDADDR_ANY);
	if (bind(s, (struct sockaddr *)&local_addr, sizeof(local_addr)) < 0) {
		BTC_DEBUG("BlueToothChannel::Open(): Can't bind socket. %s (%d)\n", strerror(errno), errno);
		close(s);
		return -1;
	}

	if ( fcntl( s, F_SETFL, fcntl( s, F_GETFL ) | O_NONBLOCK ) < 0 )
		BTC_DEBUG("BlueToothChannel::Send(): cannot set to non-blocking mode\n");

	memset(&remote_addr, 0, sizeof(remote_addr));
	remote_addr.rc_family = AF_BLUETOOTH;
	bacpy(&remote_addr.rc_bdaddr, &bdaddr);
	remote_addr.rc_channel = btChannel;

	r = connect(s, (struct sockaddr *)&remote_addr, sizeof(remote_addr));
	if ((r < 0) && (errno == EAGAIN || errno == EINPROGRESS))
	{
		r = 0;
		while (!r && !aCancel)
		{
			tv.tv_sec = 0;
			tv.tv_usec = CANCEL_POLL_INTERVAL;	
			FD_ZERO(&wfds);	FD_SET(s, &wfds);
			r = select(s + 1, NULL, &wfds, NULL, &tv);
		}
	}

	if ( aCancel )
	{
		BTC_DEBUG("BlueToothChannel::Open(): Open canceled\n");
		close(s);
		return 0;
	}

	socklen_t len = sizeof(r);
	getsockopt(s, SOL_SOCKET, SO_ERROR, (void*)&r, &len );

	if (r)
	{
		BTC_DEBUG("BlueToothChannel::Open(): Can't connect. %s (%d)\n", strerror(errno), errno);
		close(s);
		return -1;
	}
	else
	{
		btFd = s;
		BTC_DEBUG("BlueToothChannel::Open(): RFCOMM channel connected, fd = %d\n", btFd);
	}

	return 0;
}

int BluetoothChannel::Open()
{
	int dummy = 0;
	return Open(dummy);
}

int BluetoothChannel::Listen(int aChannel, volatile int& aCancel )
{

	bdaddr_t		local;
	bdaddr_t		bdaddr;
	struct sockaddr_rc 	addr;
	socklen_t		socklen;
	int 			s;

label_retry:

	hci_devba(0, &local);
	str2ba(btAddr, &bdaddr);

	if ((btFd = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) 
	{
		BTC_DEBUG("BlueToothChannel::Listen(): Can't create socket. %s (%d)\n", strerror(errno), errno);
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.rc_family = AF_BLUETOOTH;
	bacpy(&addr.rc_bdaddr, &local);
	addr.rc_channel = aChannel;
	if (bind(btFd, (struct sockaddr *)&addr, sizeof(addr)) < 0)
	{
		BTC_DEBUG("BlueToothChannel::Listen(): Can't bind socket. %s (%d)\n", strerror(errno), errno);
		close(btFd);
		return -1;
	}

	if ( fcntl( btFd, F_SETFL, fcntl( btFd, F_GETFL ) | O_NONBLOCK ) < 0 )
		BTC_DEBUG("BlueToothChannel::Listen(): cannot set to non-blocking mode\n");

	if (listen(btFd, MAX_CONNECTIONS)<0)
	{
		BTC_DEBUG("BlueToothChannel::Listen(): Can't listen socket. %s (%d)\n", strerror(errno), errno);
		close(btFd);
		return -1;
	}

	do
	{
		memset(&addr, 0, sizeof(addr));
		addr.rc_family = AF_BLUETOOTH;
		bacpy(&addr.rc_bdaddr, &bdaddr);
		addr.rc_channel = aChannel;
		usleep( CANCEL_POLL_INTERVAL );
		s = accept(btFd, (struct sockaddr *)&addr, &(socklen = sizeof(addr)));
	} 
	while( s < 0 && errno == EAGAIN && !aCancel );

	if ( aCancel )
	{
		BTC_DEBUG("BlueToothChannel::Listen(): Listen canceled\n");
		if (s > 0)
			close(s);
		close(btFd);
		btFd = -1;
		return 0;
	}

	if ( s < 0 )
	{
		BTC_DEBUG("BlueToothChannel::Listen(): Can't accept socket. %s (%d)\n", strerror(errno), errno);
		close(btFd);
		btFd = -1;
		return -1;
	}
	else if ( s > 0 )
	{
		if ( bacmp( &bdaddr, &addr.rc_bdaddr ) )
		{
			BTC_DEBUG("BlueToothChannel::Listen(): attempted connect from the different phone, REJECT.\n");
			close( s );
			close( btFd );
			goto label_retry;
		}
		close(btFd);
		btFd = s;
		if ( fcntl( btFd, F_SETFL, fcntl( btFd, F_GETFL ) | O_NONBLOCK ) < 0 )
			BTC_DEBUG("BlueToothChannel::Listen(): cannot set to non-blocking mode\n");
		BTC_DEBUG("BlueToothChannel::Listen(): RFCOMM channel accepted\n");
	}
	
	return 0;
}
int BluetoothChannel::Close()
{
	int result = 0;

	if (btFd > 0)
	{
		if ( fcntl( btFd, F_SETFL, fcntl( btFd, F_GETFL ) & ~O_NONBLOCK ) < 0 )
			BTC_DEBUG("BlueToothChannel::Close(): cannot set to non-blocking mode\n");
		result = close(btFd);
	}
	else
		result = 0;

	if (result < 0 )
	{
		BTC_DEBUG("BlueToothChannel::Close(): Closing channel failed: %s (%d)\n", strerror(errno), errno);
	}

	btFd = -1;

	return result;
}

int BluetoothChannel::Send(char* aBuffer, int aSize, volatile int& aCancel)
{
	int r;

	if ( btFd < 0 )
	{
		BTC_DEBUG("BlueToothChannel::Send(): Sending failed due to invalid file descriptor.\n");
		return -1;
	}

	while (((r = write(btFd, aBuffer, aSize) ) < 0) && !aCancel)
	{
		if (errno != EAGAIN)
			break;
		usleep(CANCEL_POLL_INTERVAL);
	}

	return r;
}

int BluetoothChannel::Receive(char* aBuffer, int* aBufferLength)
{

	if ( btFd < 0 )
	{
		BTC_DEBUG("BlueToothChannel::Receive(): Receiving failed due to invalid file descriptor.\n");
		return -1;
	}


	int		n, result;

	n = read(btFd, aBuffer, *aBufferLength );

	if ( n < 0 )
	{
		n = 0;
		if ( errno == EAGAIN )
		{
			result = 0;
		}
		else
		{
			BTC_DEBUG("BlueToochChannel::Receive(): Reading form the %s channel %d failed: %s (%d)\n", btAddr, btChannel, strerror(errno), errno);
			result = -1;
		}
	}
	else
		result = 0;

	*aBufferLength = n;

	return result;

}
/*
int	BluetoothChannel::Send(const unsigned char* aBuffer, int aBufferLength)
{
	return write(btFd, aBuffer, aBufferLength);
}
*/
int	BluetoothChannel::Receive(unsigned char* aBuffer, int aBufferLength, int aTimeout)
{
	int		nfds, n, result;
	struct timeval	timeout;
	fd_set		testfds;

	FD_ZERO(&testfds);
	FD_SET(btFd, &testfds);
    
	//do not wait
	timeout.tv_sec = aTimeout / 1000;
	timeout.tv_usec = (aTimeout - timeout.tv_sec * 1000) * 1000;
     
	//check if anything to read 
	nfds = select(FD_SETSIZE, &testfds, (fd_set *)0,(fd_set *)0, &timeout);

	n = 0;

	if ( nfds > 0 )
	{
		if(FD_ISSET(btFd,&testfds))
		{
			long nread = 1;
			//ioctl(btFd, FIONREAD, &nread); //todo: find a way to check if the read on the socket will be non-blocking

			if ( nread )
			{
				n = read(btFd, aBuffer, aBufferLength);
				if ( n < 0 )
				{
					BTC_DEBUG("BlueToothChannel::Receive(): Reading form the %s channel %d failed: %s (%d)\n", btAddr, btChannel, strerror(errno), errno);
					n = 0;
					result = -1;
				}
				else
				{
					result = 0;
				}
			}
			else
			{
				result = 0;
			}
		}
		else
		{
			//this should never happen: we are waiting for data on btFd descriptor
			BT_ASSERT(false);
		}
	}
	else if ( nfds == 0 )
	{
		result = 0;
	}
	else
	{
		BTC_DEBUG("BlueToothChannel::Receive(): Select on the %s channel %d failed: %s (%d)\n", btAddr, btChannel, strerror(errno), errno);
		result = -1;
	}
	return result;
}

int	BluetoothChannel::SendBinary(const unsigned char* aBuffer, int aBufferLength)
{
//	BT_ASSERT( devFd == -1 );
	return write(btFd, aBuffer, aBufferLength);
}

int	BluetoothChannel::ReceiveBinary(char* aBuffer, int aBufferLength, int aTimeout)
{



//	BT_ASSERT( devFd == -1 );


  int		nfds, n, result;
	struct timeval	timeout;
	fd_set		testfds;

	FD_ZERO(&testfds);
	FD_SET(btFd, &testfds);

//printf("btFd = %d\n",btFd);


	//do not wait
	timeout.tv_sec = aTimeout/1000;
	timeout.tv_usec = (aTimeout%1000)*1000;
     
	//check if anything to read 
	nfds = select(btFd+1/*FD_SETSIZE*/, &testfds, (fd_set *)0,(fd_set *)0, &timeout);

//printf("nfds = %d\n",nfds);

	n = 0;

	if ( nfds > 0 )
	{
		if(FD_ISSET(btFd,&testfds))
		{
//printf("FD_ISSET(btFd,&testfds)\n");
			n = read(btFd, aBuffer, aBufferLength);
//printf("nread = %d\n",n);
			if ( n < 0 )
			{
				BTC_DEBUG("BlueToothChannel::ReceiveBinary(): Reading form the %s channel %d failed: %s (%d)\n", btAddr, btChannel, strerror(errno), errno);
				result = 0;
			}
			else
			{
				result = n;
			}
		}
		else
		{
			//this should never happen: we are waiting for data on btFd descriptor
			BT_ASSERT(false);
		}
	}
	else if ( nfds == 0 )
	{
		result = 0;
	}
	else
	{
		result = 0;
	}

	return result;
}

int BluetoothChannel::CreateTTY(int dev_id)
{
	int ctl, ntry = 3;
	struct rfcomm_dev_req req;
	struct stat st;


	if ( dev_id == btDev )
	{
		BTC_DEBUG("BlueToothChanneli:CreateTTY(%d): reusing existing device.\n", dev_id);
		return 0;
	}

	memset(&req, 0, sizeof(req));
	req.dev_id = dev_id;
	bacpy(&req.src, BDADDR_ANY);
	str2ba(btAddr, &req.dst);
	req.channel = btChannel;

	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM)) < 0 ) {
		BTC_DEBUG("BlueToothChannel::CreateTTY(%d): Can't open RFCOMM control socket\n", dev_id);
		return -1;
	}

	if ( ioctl(ctl, RFCOMMCREATEDEV, &req) < 0 )
	{
		BTC_DEBUG("BlueToothChannel::CreateTTY(%d): Can't create device\n", dev_id);
		close( ctl );
		return -1;
	}

	sprintf(btTTY, "/dev/rfcomm%d", dev_id);
	while (stat(btTTY, &st) < 0) {
		sprintf(btTTY, "/dev/bluetooth/rfcomm/%d", dev_id);
		if (stat(btTTY, &st) < 0) {
			sprintf(btTTY, "/dev/rfcomm%d", dev_id);
			if (ntry--) {
				sleep(1);
				continue;
			}

			memset(&req, 0, sizeof(req));
			req.dev_id = dev_id;
			ioctl(ctl, RFCOMMRELEASEDEV, &req);
			sprintf(btTTY, "");
			close( ctl );
			return -1;
		}
	}

	btDev = dev_id;
	close( ctl );

	return 0;
}

char* BluetoothChannel::GetTTY()
{
	return btTTY;
}
bool BluetoothChannel::IsTTYBusy()
{
	static bool rfcomm_state[] = {
		true,//"unknown",
		true,//"connected",
		false,//"clean",
		true,//"bound",
		true,//"listening",
		true,//"connecting",
		true,//"connecting",
		true,//"config",
		true,//"disconnecting",
		false//"closed"
	};

	int ctl, i;
	bool result;
	struct rfcomm_dev_list_req *dl;
	struct rfcomm_dev_info *di;

	if ( btDev < 0 )
	{
		BTC_DEBUG("WARNING: BlueToothChannel::IsTTYBusy(): device does not exist.\n");
		return -1;
	}

	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM)) < 0 ) {
		BTC_DEBUG("BlueToothChannel::IsTTYBusy(): Can't open RFCOMM control socket");
		return -1;
	}

	dl = (rfcomm_dev_list_req*)malloc(sizeof(*dl) + RFCOMM_MAX_DEV * sizeof(*di));
	if (!dl) {
		BTC_DEBUG("BlueToothChannel::IsTTYBusy(): Can't allocate memory");
		close( ctl );
		return -1;
	}

	dl->dev_num = RFCOMM_MAX_DEV;
	di = dl->dev_info;
	if (ioctl(ctl, RFCOMMGETDEVLIST, (void *) dl) < 0) {
		BTC_DEBUG("BlueToothChannel::IsTTYBusy(): Can't get device list");
		close( ctl );
		return -1;
	}

	printf("dev_num = %d\n", dl->dev_num);
	result = false;
	for (i = 0; i < dl->dev_num; i++)
	{
//		if ( di->id == btDev )
		{
			printf("dev %d, state = %d\n", di->id, (di + i)->state);
			result = rfcomm_state[(di + i)->state];
			break;
		}
	}

	free( dl );
	close( ctl );

	return result;
}

int BluetoothChannel::ReleaseTTY()
{
	int ctl;
	struct rfcomm_dev_req req;

	if ( btDev < 0 )
	{
		BTC_DEBUG("WARNING: BlueToothChannel::ReleaseTTY(): device does not exist.\n");
		return 0;
	}

	memset(&req, 0, sizeof(req));
	req.dev_id = btDev;

	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM)) < 0 ) {
		BTC_DEBUG("BlueToothChannel::ReleaseTTY(): Can't open RFCOMM control socket");
		return -1;
	}

	if ( ioctl(ctl, RFCOMMRELEASEDEV, &req)  < 0 )
	{
		BTC_DEBUG("BlueToothChannel::ReleaseTTY(): Can't release device");
		close( ctl );
		return -1;
	}

	btDev = -1;
	sprintf(btTTY, "");
	close( ctl );

	return 0;
}

char* BluetoothChannel::GetAddress()
{
	return btAddr;
}

int BluetoothChannel::CmpAddress(char* aAddress)
{
	bdaddr_t addr1, addr2;
	if (!aAddress)
		return -1;
	str2ba(btAddr, &addr1); 
	str2ba(aAddress, &addr2);
	return bacmp(&addr1, &addr2);
}

int BluetoothChannel::GetChannel()
{
	return btChannel;
}

int BluetoothChannel::IsAclLinkActive()
{

	struct hci_dev_list_req *dl = NULL;
	struct hci_dev_req *dr;
	struct hci_conn_list_req *cl = NULL;
	struct hci_conn_info *ci;
	bdaddr_t bdaddr;
	int ii, i, sk;
	int result = -1;

	sk = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI);
	if (sk < 0)
		return -1;

	dl = (struct hci_dev_list_req*)malloc(HCI_MAX_DEV * sizeof(*dr) + sizeof(*dl));
	if (!dl)
		goto done;
	if (!(cl = (struct hci_conn_list_req*)malloc(10 * sizeof(*ci) + sizeof(*cl))))
		goto done;

	dl->dev_num = HCI_MAX_DEV;
	dr = dl->dev_req;

	if (ioctl(sk, HCIGETDEVLIST, (void *) dl))
	{
		BTC_DEBUG("BlueToothChannel::IsAclLinkActive(): Can't get device list");
		goto done;
	}

	for (i = 0; i < dl->dev_num; i++, dr++)
	{
		if (hci_test_bit(HCI_UP, &dr->dev_opt))
		{
			cl->dev_id = dr->dev_id;
			cl->conn_num = 10;
			ci = cl->conn_info;

			if (ioctl(sk, HCIGETCONNLIST, (void *) cl))
			{
				BTC_DEBUG("BlueToothChannel::IsAclLinkActive(): Can't get connection list");
				goto done;
			}

			for (ii = 0; ii < cl->conn_num; ii++, ci++)
			{
				str2ba(btAddr, &bdaddr);
				if ( !bacmp(&ci->bdaddr, &bdaddr) && ci->type == ACL_LINK)
				{
					result = dr->dev_id;
					goto done;
				}
			}
		}
	}

done:
	close(sk);
	if (cl)
		free(cl);
	if (dl) 
		free(dl);
	return result;
}
void BluetoothChannel::DestroyAclLink()
{
	struct hci_conn_info_req *cr;
	bdaddr_t bdaddr;
	int dd, dev_id = IsAclLinkActive();

	if (dev_id < 0)
		return;

	dd = hci_open_dev(dev_id);
	if (dd < 0) {
		BTC_DEBUG("BlueToothChannel::DestroyAclLink(): HCI device open failed");
		return;
	}

	cr = (struct hci_conn_info_req*)malloc(sizeof(*cr) + sizeof(struct hci_conn_info));
	if (!cr)
		return;
	str2ba(btAddr, &bdaddr);
	bacpy(&cr->bdaddr, &bdaddr);
	cr->type = ACL_LINK;
	if (ioctl(dd, HCIGETCONNINFO, (unsigned long) cr) < 0) {
		BTC_DEBUG("BlueToothChannel::DestroyAclLink(): Get connection info failed");
		return;
	}

	if (hci_disconnect(dd, htobs(cr->conn_info->handle),
				HCI_OE_USER_ENDED_CONNECTION, 10000) < 0)
		BTC_DEBUG("BlueToothChannel::DestroyAclLink(): Disconnect failed");

	close(dd);
	free(cr);

	return;

}
